@@ -193,6 +193,29 @@ static const struct ieee80211_iface_combination rsi_iface_combinations[] = {
 };
 #endif
 
+#ifdef CONFIG_CARACALLA_BOARD
+struct reg_map {
+	char country_code[2];
+	u8 region_code; 
+};
+
+static struct reg_map rsi_caracalla_reg_db[MAX_REG_COUNTRIES] = {
+	{"AU", NL80211_DFS_ETSI}, {"AT", NL80211_DFS_ETSI},
+	{"BE", NL80211_DFS_ETSI}, {"BR", NL80211_DFS_WORLD},
+	{"CA", NL80211_DFS_FCC}, {"CL", NL80211_DFS_WORLD},
+	{"CN", NL80211_DFS_WORLD}, {"CO", NL80211_DFS_FCC},
+	{"CZ", NL80211_DFS_ETSI}, {"DK", NL80211_DFS_ETSI},
+	{"FI", NL80211_DFS_ETSI}, {"FR", NL80211_DFS_ETSI},
+	{"DE", NL80211_DFS_ETSI}, {"HK", NL80211_DFS_WORLD},
+	{"IN", NL80211_DFS_WORLD}, {"ID", NL80211_DFS_WORLD},
+	{"IE", NL80211_DFS_ETSI}, {"IL", NL80211_DFS_ETSI},
+	{"IT", NL80211_DFS_ETSI}, {"JP", NL80211_DFS_JP},
+	{"KR", NL80211_DFS_WORLD}, {"LU", NL80211_DFS_ETSI},
+	{"MY", NL80211_DFS_WORLD}, {"MX", NL80211_DFS_FCC},
+	{"MA", NL80211_DFS_WORLD}, {"NL", NL80211_DFS_ETSI},
+};
+#endif
+
 struct ieee80211_vif *rsi_get_vif(struct rsi_hw *adapter, u8 *mac)
 {
 	u8 i;
@@ -337,6 +360,9 @@ static int rsi_mac80211_hw_scan_start(struct ieee80211_hw *hw,
 
 	ven_rsi_dbg(INFO_ZONE, "***** Hardware scan start *****\n");
 
+	if (common->fsm_state != FSM_MAC_INIT_DONE)
+		return -ENODEV;
+
 	if (scan_req->n_channels == 0)
 		return -EINVAL;
 
@@ -430,6 +456,9 @@ static void rsi_mac80211_hw_scan_cancel(struct ieee80211_hw *hw,
 void ven_rsi_mac80211_detach(struct rsi_hw *adapter)
 {
 	struct ieee80211_hw *hw = adapter->hw;
+#ifdef CONFIG_HW_SCAN_OFFLOAD
+	struct rsi_common *common = adapter->priv;
+#endif
 #if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 7, 0))
 	enum nl80211_band band;
 #else
@@ -438,11 +467,15 @@ void ven_rsi_mac80211_detach(struct rsi_hw *adapter)
 
 	ven_rsi_dbg(INFO_ZONE, "Detach mac80211...\n");
 
+#ifdef CONFIG_HW_SCAN_OFFLOAD
+	flush_workqueue(common->scan_workqueue);
+#endif
 	if (hw) {
 		ieee80211_stop_queues(hw);
 		ieee80211_unregister_hw(hw);
 		ieee80211_free_hw(hw);
 		adapter->hw = NULL;
+		adapter->sc_nvifs = 0;
 	}
 
 #if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 7, 0))
@@ -513,7 +546,7 @@ static void rsi_mac80211_tx(struct ieee80211_hw *hw,
   struct ieee80211_vif *vif = adapter->vifs[adapter->sc_nvifs - 1];
 
 #ifdef CONFIG_VEN_RSI_WOW
-	if (common->suspend_flag) {
+	if (common->wow_flags & RSI_WOW_ENABLED) {
 		ieee80211_free_txskb(common->priv->hw, skb);
 		return;
 	}
@@ -580,9 +613,7 @@ static void rsi_mac80211_stop(struct ieee80211_hw *hw)
 {
 	struct rsi_hw *adapter = hw->priv;
 	struct rsi_common *common = adapter->priv;
-#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 9, 0))
-	struct cfg80211_scan_info info;
-#endif
+
 	ven_rsi_dbg(ERR_ZONE, "===> Interface DOWN <===\n");
 
 	mutex_lock(&common->mutex);
@@ -596,7 +627,8 @@ static void rsi_mac80211_stop(struct ieee80211_hw *hw)
 	wiphy_rfkill_stop_polling(hw->wiphy);
 
 	/* Block all rx frames */
-	rsi_send_rx_filter_frame(common, 0xffff);
+	if (common->fsm_state == FSM_MAC_INIT_DONE)
+		rsi_send_rx_filter_frame(common, 0xffff);
 	
 	mutex_unlock(&common->mutex);
 }
@@ -1001,6 +1033,9 @@ static int rsi_mac80211_config(struct ieee80211_hw *hw,
 	struct ieee80211_conf *conf = &hw->conf;
 	int status = -EOPNOTSUPP;
 
+	if (common->fsm_state != FSM_MAC_INIT_DONE)
+		return -ENODEV;
+
 	mutex_lock(&common->mutex);
 
 	/* channel */
@@ -1139,6 +1174,9 @@ static void rsi_mac80211_bss_info_changed(struct ieee80211_hw *hw,
 	ven_rsi_dbg(INFO_ZONE, "%s: BSS status changed; changed=%08x\n",
 		__func__, changed);
 
+	if (common->fsm_state != FSM_MAC_INIT_DONE)
+		return;
+
 	mutex_lock(&common->mutex);
 
 	if ((changed & BSS_CHANGED_ASSOC) &&
@@ -1176,10 +1214,13 @@ static void rsi_mac80211_bss_info_changed(struct ieee80211_hw *hw,
 			bss->dtim_period);
 
 		/* If UAPSD is updated send ps params */
-		if (common->uapsd_bitmap) {
-			ven_rsi_dbg(INFO_ZONE, "Configuring UAPSD\n");
-			rsi_conf_uapsd(adapter);
-		}
+		if (bss->assoc) {
+			if (common->uapsd_bitmap) {
+				ven_rsi_dbg(INFO_ZONE, "Configuring UAPSD\n");
+				rsi_conf_uapsd(adapter);
+			}
+		} else
+			common->uapsd_bitmap = 0;
 	}
 
 	if ((vif->type == NL80211_IFTYPE_STATION) &&
@@ -1300,6 +1341,8 @@ static int rsi_mac80211_conf_tx(struct ieee80211_hw *hw,
 
 	if (params->uapsd)
 		common->uapsd_bitmap |= idx;
+	else
+		common->uapsd_bitmap &= (~idx);
 
 	mutex_unlock(&common->mutex);
 	return 0;
@@ -1418,6 +1461,8 @@ static int rsi_mac80211_set_key(struct ieee80211_hw *hw,
 		break;
 
 	case DISABLE_KEY:
+		if ((vif->type == NL80211_IFTYPE_STATION) ||
+		    (vif->type == NL80211_IFTYPE_P2P_CLIENT))
 		secinfo->security_enable = false;
 		ven_rsi_dbg(ERR_ZONE, "%s: RSI del key\n", __func__);
 		memset(key, 0, sizeof(struct ieee80211_key_conf));
@@ -1447,6 +1492,7 @@ static int rsi_mac80211_set_key(struct ieee80211_hw *hw,
  *
  * Return: status: 0 on success, negative error code on failure.
  */
+#ifndef CONFIG_CARACALLA_BOARD
 #if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 4, 0))
 static int rsi_mac80211_ampdu_action(struct ieee80211_hw *hw,
 				     struct ieee80211_vif *vif,
@@ -1469,6 +1515,12 @@ static int rsi_mac80211_ampdu_action(struct ieee80211_hw *hw,
 				     struct ieee80211_vif *vif,
 				     struct ieee80211_ampdu_params *params)
 #endif
+#else
+static int rsi_mac80211_ampdu_action(struct ieee80211_hw *hw,
+				     struct ieee80211_vif *vif,
+				     struct ieee80211_ampdu_params *params)
+#endif
+	
 {
 	int status = 1;
 	struct rsi_hw *adapter = hw->priv;
@@ -1478,7 +1530,7 @@ static int rsi_mac80211_ampdu_action(struct ieee80211_hw *hw,
 	u8 ii = 0;
 	u8 sta_id = 0;
 
-#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 6, 0))
+#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(4, 6, 0)) || defined(CONFIG_CARACALLA_BOARD))
 	u16 tid = params->tid;
 	u8 buf_size = params->buf_size;
 	enum ieee80211_ampdu_mlme_action action = params->action;
@@ -1492,7 +1544,7 @@ static int rsi_mac80211_ampdu_action(struct ieee80211_hw *hw,
 
 	mutex_lock(&common->mutex);
 
-#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 6, 0))
+#if ((LINUX_VERSION_CODE < KERNEL_VERSION(4, 6, 0)) && !(defined(CONFIG_CARACALLA_BOARD)))
 	if (ssn != NULL)
 		seq_no = *ssn;
 #else
@@ -1800,7 +1852,7 @@ static int rsi_mac80211_sta_add(struct ieee80211_hw *hw,
 	struct rsi_common *common = adapter->priv;
 	bool sta_exist = 0;
 
-	rsi_hex_dump(INFO_ZONE, "Station Add: ", sta->addr, ETH_ALEN);
+	rsi_hex_dump(INFO_ZONE, "Station Add", sta->addr, ETH_ALEN);
 
 	mutex_lock(&common->mutex);
 
@@ -1853,12 +1905,16 @@ static int rsi_mac80211_sta_add(struct ieee80211_hw *hw,
 			}
 			for (j = 0; j < IEEE80211_NUM_ACS; j++)
 				common->stations[i].seq_no[j] = 1;
+			for (j = 0; j < IEEE80211_NUM_TIDS; j++)
+				common->stations[i].start_tx_aggr[j] = false;
 			common->num_stations++;
 		} else {
 			common->stations[i].sta = sta;
 			common->stations[i].sta_id = i;
 			for (j = 0; j < IEEE80211_NUM_ACS; j++)
 				common->stations[i].seq_no[j] = 1;
+			for (j = 0; j < IEEE80211_NUM_TIDS; j++)
+				common->stations[i].start_tx_aggr[j] = false;
 		}
 	}
 
@@ -1879,7 +1935,9 @@ static int rsi_mac80211_sta_add(struct ieee80211_hw *hw,
 		}
 	}
 
-	if (sta->ht_cap.ht_supported) {
+	if (((vif->type == NL80211_IFTYPE_STATION) ||
+	     (vif->type == NL80211_IFTYPE_P2P_CLIENT)) &&
+	    (sta->ht_cap.ht_supported)) {
 		common->vif_info[0].is_ht = true;
 		common->bitrate_mask[NL80211_BAND_2GHZ] =
 				sta->supp_rates[NL80211_BAND_2GHZ];
@@ -1910,9 +1968,8 @@ static int rsi_mac80211_sta_remove(struct ieee80211_hw *hw,
 	struct rsi_hw *adapter = hw->priv;
 	struct rsi_common *common = adapter->priv;
 	struct ieee80211_bss_conf *bss = &vif->bss_conf;
-//		&adapter->vifs[adapter->sc_nvifs - 1]->bss_conf;
 
-	rsi_hex_dump(INFO_ZONE, "Station Removed: ", sta->addr, ETH_ALEN);
+	rsi_hex_dump(INFO_ZONE, "Station Removed", sta->addr, ETH_ALEN);
 
 	mutex_lock(&common->mutex);
 	if ((vif->type == NL80211_IFTYPE_AP) ||
@@ -1978,6 +2035,9 @@ static void rsi_mac80211_sw_scan_start(struct ieee80211_hw *hw,
 	struct rsi_common *common = adapter->priv;
 	struct ieee80211_bss_conf *bss = &adapter->vifs[0]->bss_conf;
 
+	if (common->fsm_state != FSM_MAC_INIT_DONE)
+		return;
+
 	if (common->p2p_enabled)
 		return;
 	if (!bss->assoc)
@@ -2102,7 +2162,7 @@ static int rsi_mac80211_get_antenna(struct ieee80211_hw *hw,
 	return 0;	
 }
 
-static const char *regdfs_region_str(enum nl80211_dfs_regions dfs_region)
+static const char *regdfs_region_str(u8 dfs_region)
 {
         switch (dfs_region) {
         case NL80211_DFS_UNSET:
@@ -2113,6 +2173,8 @@ static const char *regdfs_region_str(enum nl80211_dfs_regions dfs_region)
                 return "ETSI";
         case NL80211_DFS_JP:
                 return "JP";
+	case NL80211_DFS_WORLD:
+		return "WORLD";
         }
         return "Unknown";
 }
@@ -2129,14 +2191,32 @@ static void rsi_reg_notify(struct wiphy *wiphy,
 	struct ieee80211_vif *vif = adapter->vifs[adapter->sc_nvifs -1];
 #endif
 	int i;
+	u8 region_code;
 
 	mutex_lock(&common->mutex);
 
+	ven_rsi_dbg(INFO_ZONE,
+		"%s: country = %s dfs_region = %d\n",
+		__func__, request->alpha2, request->dfs_region);
+
+	region_code = request->dfs_region;
+#ifdef CONFIG_CARACALLA_BOARD
+	for (i = 0; i < ARRAY_SIZE(rsi_caracalla_reg_db); i++) {
+		if (!memcmp(rsi_caracalla_reg_db[i].country_code,
+		            request->alpha2, 2)) {
+			region_code = rsi_caracalla_reg_db[i].region_code;
+			break;
+		}
+	}
+	if (!memcmp(request->alpha2, "00", 2))
+		region_code = NL80211_DFS_JP;
+#endif
+
 	ven_rsi_dbg(ERR_ZONE, "%s: Updating regulatory for region %s\n",
-		__func__, regdfs_region_str(request->dfs_region));
+		__func__, regdfs_region_str(region_code));
 
 	sband = wiphy->bands[NL80211_BAND_2GHZ];
-	switch (request->dfs_region) {
+	switch (region_code) {
 	case NL80211_DFS_FCC:
 		for(i = 0; i < sband->n_channels; i++){
 	 		ch = &sband->channels[i];
@@ -2163,7 +2243,7 @@ static void rsi_reg_notify(struct wiphy *wiphy,
 
 	case NL80211_DFS_UNSET:
 	case NL80211_DFS_ETSI:
-		
+	case NL80211_DFS_WORLD:
 		for(i = 0; i < sband->n_channels; i++){
 	 		ch = &sband->channels[i];
 			
@@ -2201,7 +2281,7 @@ static void rsi_reg_notify(struct wiphy *wiphy,
 	default:
 		ven_rsi_dbg(ERR_ZONE,
 			"Wrong Country region code seleted\ndfs_region = %d\n",
-			request->dfs_region);
+			region_code);
 		break;
 	}
 
@@ -2217,15 +2297,11 @@ static void rsi_reg_notify(struct wiphy *wiphy,
 				ch->flags |= IEEE80211_CHAN_NO_IR;
 		}
 	}
-	
-	ven_rsi_dbg(INFO_ZONE,
-		"country = %s dfs_region = %d\n",
-		request->alpha2, request->dfs_region);
 
 #ifndef CONFIG_HW_SCAN_OFFLOAD 
 	/* If DFS region or country is changed configure back ground scan
 	 * params to device again */
-	if ((adapter->dfs_region != request->dfs_region) ||
+	if ((adapter->dfs_region != region_code) ||
 	    (memcmp(adapter->country, request->alpha2, 2))) {
 		if (common->bgscan_en) {
 			rsi_send_bgscan_params(common, 0);
@@ -2237,7 +2313,15 @@ static void rsi_reg_notify(struct wiphy *wiphy,
 	}
 #endif
 	
-	adapter->dfs_region = request->dfs_region;
+	if (region_code == NL80211_DFS_UNSET)
+		region_code = NL80211_DFS_WORLD;
+	ven_rsi_dbg(INFO_ZONE, "Region code: %s\n", regdfs_region_str(region_code));
+
+	/* Firmware region code values 1 less than the standard
+	 * linux values; At this point DFS_UNSET is not set
+	 * as this is set to WORLD (4)
+	 */
+	adapter->dfs_region = region_code - 1;
 	adapter->country[0] = request->alpha2[0];
 	adapter->country[1] = request->alpha2[1];
 	mutex_unlock(&common->mutex);
@@ -2293,42 +2377,34 @@ static u16 rsi_wow_map_triggers(struct rsi_common *common,
 }
 #endif
 
-#ifdef CONFIG_PM
-int rsi_mac80211_suspend(struct ieee80211_hw *hw,
-			 struct cfg80211_wowlan *wowlan)
-{
 #ifdef CONFIG_VEN_RSI_WOW
-	struct rsi_hw *adapter = hw->priv;
+int rsi_config_wowlan(struct rsi_hw *adapter, struct cfg80211_wowlan *wowlan)
+{
 	struct rsi_common *common = adapter->priv;
 	u16 triggers = 0;
 	u16 rx_filter_word = 0;
 	struct ieee80211_bss_conf *bss = &adapter->vifs[0]->bss_conf;
-#endif
-	int ret = 0;
 
-	ven_rsi_dbg(INFO_ZONE, "***** mac80211 suspend called ******\n");
+	ven_rsi_dbg(INFO_ZONE, "Config WoWLAN to device\n");
 
-#ifdef CONFIG_VEN_RSI_WOW
 	if (WARN_ON(!wowlan)) {
-		ven_rsi_dbg(ERR_ZONE,
-			"##### WoW triggers not enabled #####\n");
-		ret = -EINVAL;
-		goto fail_wow;
+		ven_rsi_dbg(ERR_ZONE, "WoW triggers not enabled\n");
+		return -EINVAL;
 	}
 
 	triggers = rsi_wow_map_triggers(common, wowlan);
 	if (!triggers) {
 		ven_rsi_dbg(ERR_ZONE, "%s:No valid WoW triggers\n",__func__);
-		ret = -EINVAL;
-		goto fail_wow;
+		return -EINVAL;
 	}
 	if (!bss->assoc) {
 		ven_rsi_dbg(ERR_ZONE,
 			"Cannot configure WoWLAN (Station not connected)\n");
-		common->suspend_flag = STATION_NOT_CONNECTED;
-		ret = 0;
-		goto fail_wow;
+		common->wow_flags |= RSI_WOW_NO_CONNECTION;
+		return 0;
 	}
+	ven_rsi_dbg(INFO_ZONE, "TRIGGERS %x\n", triggers);
+	rsi_send_wowlan_request(common, triggers, 1);
 
 	/* Send updated vap caps */
 	rsi_send_vap_dynamic_update(common);
@@ -2336,43 +2412,65 @@ int rsi_mac80211_suspend(struct ieee80211_hw *hw,
 	rx_filter_word = (ALLOW_DATA_ASSOC_PEER |
 			  DISALLOW_BEACONS |
 			  0);
-
-
-	rsi_send_wowlan_request(common, triggers, 1);
-	ven_rsi_dbg(INFO_ZONE, "TRIGGERS %x\n", triggers);
 	rsi_send_rx_filter_frame(common, rx_filter_word);
 
-        common->suspend_flag = 1;
-fail_wow:
-#endif
-        return (ret ? 1 : 0);
+        common->wow_flags |= RSI_WOW_ENABLED;
+        
+	return 0;
 }
+EXPORT_SYMBOL_GPL(rsi_config_wowlan);
+#endif
 
-static int rsi_mac80211_resume(struct ieee80211_hw *hw)
+#ifdef CONFIG_PM
+static int rsi_mac80211_suspend(struct ieee80211_hw *hw,
+				struct cfg80211_wowlan *wowlan)
 {
 #ifdef CONFIG_VEN_RSI_WOW
 	struct rsi_hw *adapter = hw->priv;
 	struct rsi_common *common = adapter->priv;
-	u16 rx_filter_word = 0;
 
-	adapter->priv->suspend_flag = 0;
+	mutex_lock(&common->mutex);
+	if (rsi_config_wowlan(adapter, wowlan)) {
+		ven_rsi_dbg(ERR_ZONE, "Failed to configure WoWLAN\n");
+		mutex_unlock(&common->mutex);
+		return 1;
+	}
+	mutex_unlock(&common->mutex);
+#endif
+	return 0;
+}
+
+static int rsi_mac80211_resume(struct ieee80211_hw *hw)
+{
+	struct rsi_hw *adapter = hw->priv;
+	struct rsi_common *common = adapter->priv;
+#ifdef CONFIG_VEN_RSI_WOW
+	u16 rx_filter_word = 0;
+        
+	adapter->priv->wow_flags = 0;
 #endif
 	
 	ven_rsi_dbg(INFO_ZONE, "%s: mac80211 resume\n", __func__);
 
+	if (common->hibernate_resume)
+		return 0;
+
 #ifdef CONFIG_VEN_RSI_WOW
+	mutex_lock(&common->mutex);
 	rsi_send_wowlan_request(common, 0, 0);
-	//rx_filter_word = 0xE ;
 
 	rx_filter_word = (ALLOW_DATA_ASSOC_PEER |
 			  ALLOW_CTRL_ASSOC_PEER |
 			  ALLOW_MGMT_ASSOC_PEER |
 			  0);
 	rsi_send_rx_filter_frame(common, rx_filter_word);
+	mutex_unlock(&common->mutex);
 #endif
+
 	return 0;
 }
 #endif
+
 char *rsi_vif_type_to_name(enum nl80211_iftype vif_type)
 {
 	switch (vif_type) {
@@ -2665,8 +2763,8 @@ int rsi_mac80211_attach(struct rsi_common *common)
 	hw->max_rate_tries = MAX_RETRIES;
 	hw->uapsd_queues = IEEE80211_MARKALL_UAPSD_QUEUES;
 	hw->uapsd_max_sp_len = IEEE80211_WMM_IE_STA_QOSINFO_SP_ALL;
-//	hw->max_tx_aggregation_subframes = 6;
-	hw->max_tx_aggregation_subframes = 4;
+	hw->max_tx_aggregation_subframes = 8;
+	hw->max_rx_aggregation_subframes = 8;
 
 	rsi_register_rates_channels(adapter, NL80211_BAND_2GHZ);
 	wiphy->bands[NL80211_BAND_2GHZ] =
@@ -2714,6 +2812,7 @@ int rsi_mac80211_attach(struct rsi_common *common)
 	/* AP Parameters */
 	wiphy->flags = WIPHY_FLAG_REPORTS_OBSS;
 	wiphy->flags |= WIPHY_FLAG_AP_UAPSD;
+	wiphy->features |= NL80211_FEATURE_INACTIVITY_TIMER;
 
 	/*wiphy->regulatory_flags = (REGULATORY_STRICT_REG |
 				   REGULATORY_CUSTOM_REG);
@@ -2732,11 +2831,13 @@ int rsi_mac80211_attach(struct rsi_common *common)
 	hw->max_listen_interval = 10;
 	wiphy->iface_combinations = rsi_iface_combinations;
 	wiphy->n_iface_combinations = ARRAY_SIZE(rsi_iface_combinations);
-	wiphy->features |= NL80211_FEATURE_INACTIVITY_TIMER;
 //	wiphy->features |= (NL80211_FEATURE_P2P_GO_CTWIN |
 //			    NL80211_FEATURE_P2P_GO_OPPPS);
 #endif
 
+	if (common->coex_mode > 1)
+		wiphy->flags |= WIPHY_FLAG_PS_ON_BY_DEFAULT;
+
 	status = ieee80211_register_hw(hw);
 	if (status) {
 		ven_rsi_dbg(ERR_ZONE, "Failed to register to mac80211\n");
